"""Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from traceutils.logger import Logger


class Oracle:
    """Traces checking engine."""

    def __init__(self):
        """Default constructor."""

        # Mapping from traces to related records.
        self._trace_rec_map = {}

    def register_record(self, record, trace):
        """Registers log record of the specified trace in this oracle."""
        if record is None:
            return

        traces_with_record = Oracle.appear_in_other_traces(self, record, trace)

        if len(traces_with_record):
            # If the record was in other traces, then it should be removed from everywhere
            for rec_trace in traces_with_record:
                self._trace_rec_map.get(rec_trace).remove(record)

        else:
            if trace not in self._trace_rec_map:
                self._trace_rec_map[trace] = []

            # Check whether trace records are ordered by time
            record_list = self._trace_rec_map[trace]
            trace_size = len(record_list)
            if trace_size > 0:
                last_record = record_list[trace_size - 1]
                if last_record.get_time() > record.get_time():
                    Logger.print_msg('Error: Can\'t register a record with time %s; the last record has %s'
                                     % (record.get_time(), last_record.get_time()))
                    return

            self._trace_rec_map[trace].append(record)

    def appear_in_other_traces(self, record, trace):
        """Returns the list of traces, other than the specified, that contain the specified record."""
        traces_with_record = []
        for map_trace, map_records in self._trace_rec_map.iteritems():
            if Oracle.contain_record(map_records, record) and (map_trace.name() != trace.name()):
                traces_with_record.append(map_trace)

        return traces_with_record

    def future_record_for_other_traces(self, record, trace):
        """Checks whether the specified record from the specified trace is a 'future' one for other traces."""
        for map_trace, map_records in self._trace_rec_map.iteritems():
            if map_trace.name() == trace.name():
                continue
            rec_len = len(map_records)
            if rec_len == 0:
                continue
            last = map_records[rec_len - 1]
            if last.get_time() >= record.get_time():
                return False
        return True

    @staticmethod
    def contain_record(ordered_by_time_records, record):
        """Checks whether the specified collection of ordered by time records contains the specified one."""
        result = False

        for item in reversed(ordered_by_time_records):

            if record.get_time() > item.get_time():
                break
            elif item == record:
                result = True
                break

        return result

    def check(self):
        """Checks whether traces are equivalent."""
        all_traces_equivalent = True

        for records in self._trace_rec_map.values():
            if len(records) != 0:
                all_traces_equivalent = False
                break

        if all_traces_equivalent:
            print 'All the traces are equivalent.'
        else:
            Logger.non_match_rec_hdr()
            for trace, records in self._trace_rec_map.iteritems():
                Logger.print_msg('\nTrace=%s\n' % trace.name())
                Logger.print_msg('%s' % '\n'.join([record.to_string() for record in records]))
